{% load i18n %}
{% load humanize %}

<div role="tabpanel" class="tab-pane" id="skill_sets">
    <h3>{% translate 'Skill Sets' %}</h3>
    <div class="table-responsive">
        <table class="table table-striped table-width-fix" id="tab_skill_sets">
            <thead>
                <tr>
                    <th>{% translate 'Skill Set Group' %}</th>
                    <th>{% translate 'Skill Set' %}</th>
                    <th>{% translate 'Has all required?' %}</th>
                    <th>{% translate 'Missing Required Skills' %}</th>
                    <th>{% translate 'Missing Recommended Skills' %}</th>
                    <th></th>
                </tr>
            </thead>

            <tbody></tbody>
        </table>
    </div>
    {% include 'memberaudit/partials/character_viewer/tabs/update_status.html' with update_section=sections_update_status.skill_sets %}
</div>
