{% extends 'admin/base_site.html' %}
{% load i18n %}

{% block content %}
<form action="" method="post">
    {% csrf_token %}
    <div>
        <ul>
        {% for obj in queryset %}
            <input type="hidden" name="_selected_action" value="{{ obj.pk }}" />
            <li>{{ obj }} </li>
        {% endfor %}
        </ul>
        <br>
        <input type="hidden" name="action" value="{{ action }}" />
        <input type="submit" name="apply" value="{% translate 'Delete' %}" style="background-color: rgb(201, 48, 44)"/>
        <a href="./"><input type="button" name="Cancel" value="{% translate 'Cancel' %}"></a>
    </div>
</form>
{% endblock %}
