# Generated by Django 4.0.7 on 2022-10-21 17:13

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0002_add_mining_ledger"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    (
                        "basic_access",
                        "Can access this app, register, and view own characters",
                    ),
                    ("share_characters", "Can share his/her characters"),
                    ("finder_access", "Can access character finder feature"),
                    ("reports_access", "Can access reports feature"),
                    ("characters_access", "Can view characters owned by others"),
                    ("exports_access", "Can access data exports"),
                    ("view_shared_characters", "Can view shared characters"),
                    ("view_same_corporation", "Can view corporation characters"),
                    ("view_same_alliance", "Can view alliance characters"),
                    ("view_everything", "Can view all characters"),
                    (
                        "notified_on_character_removal",
                        "Notified when member drops character",
                    ),
                ),
            },
        ),
    ]
