{% extends 'allianceauth/base.html' %}

{% block page_title %}
    {{ page_title|striptags }} - {{ app_title }}
{% endblock %}

{% block content %}
    <br />
    <div class="aa-memberaudit">
        {% include 'memberaudit/partials/menu.html' %}
        <h2 class="text-left">{{ page_title }}</h2>
        {% block details %}{% endblock %}
    </div>
{% endblock %}
