# Generated by Django 4.0.10 on 2023-04-19 20:03

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0006_add_localizations"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="character",
            options={
                "default_permissions": (),
                "verbose_name": "character",
                "verbose_name_plural": "characters",
            },
        ),
        migrations.AlterModelOptions(
            name="characterupdatestatus",
            options={
                "default_permissions": (),
                "verbose_name": "character update status",
                "verbose_name_plural": "character update status",
            },
        ),
        migrations.AlterModelOptions(
            name="compliancegroupdesignation",
            options={
                "default_permissions": (),
                "verbose_name": "compliance group designation",
                "verbose_name_plural": "compliance group designations",
            },
        ),
        migrations.AlterModelOptions(
            name="location",
            options={
                "default_permissions": (),
                "verbose_name": "location",
                "verbose_name_plural": "locations",
            },
        ),
        migrations.AlterModelOptions(
            name="skillset",
            options={
                "default_permissions": (),
                "verbose_name": "skill set",
                "verbose_name_plural": "skill sets",
            },
        ),
        migrations.AlterModelOptions(
            name="skillsetgroup",
            options={
                "default_permissions": (),
                "verbose_name": "skill set group",
                "verbose_name_plural": "skill set groups",
            },
        ),
        migrations.AlterModelOptions(
            name="skillsetskill",
            options={
                "default_permissions": (),
                "verbose_name": "skill",
                "verbose_name_plural": "skills",
            },
        ),
    ]
